Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Text

''' <summary>
''' Provides a generic collection that supports data binding.  
''' It supports both complex and simple scenarios when binding 
''' to a data source and it provides advanced sorting and filtering capabilities.
''' </summary>
''' <typeparam name="T"></typeparam>
<Serializable()> _
Public Class BindingListView(Of T)
   Inherits BindingList(Of T)
   Implements IBindingListView
   Implements IRaiseItemChangedEvents

   Private mSorted As Boolean = False
   Private mFiltered As Boolean = False
   Private mFilterString As String = Nothing
   Private mSortDirection As ListSortDirection = ListSortDirection.Ascending
   <NonSerialized()> Private mSortProperty As PropertyDescriptor = Nothing
   <NonSerialized()> Private mSortDescriptions As New ListSortDescriptionCollection()
   Private mOriginalCollection As New List(Of T)()

   ''' <summary>
   ''' Constructor
   ''' </summary>
   Public Sub New()
      MyBase.New()
   End Sub

   ''' <summary>
   ''' Constructor
   ''' </summary>
   ''' <param name="list">System.Collections.Generic.IList of items to be contained in the System.ComponentModel.BindingList(Of T).</param>
   Public Sub New(ByVal list As List(Of T))
      MyBase.New(list)
   End Sub

   ''' <summary>
   ''' Gets the SupportsSearchingCore flag
   ''' </summary>
   Protected Overloads Overrides ReadOnly Property SupportsSearchingCore() As Boolean
      Get
         Return True
      End Get
   End Property

   ''' <summary>
   ''' Gets the index of the row that has the given System.ComponentModel.PropertyDescriptor
   ''' </summary>
   ''' <param name="prop">Property Descriptor</param>
   ''' <param name="key">Key</param>
   Protected Overloads Overrides Function FindCore(ByVal [prop] As PropertyDescriptor, ByVal key As Object) As Integer
      For i As Integer = 0 To Count - 1
         Dim item As T = Me(i)
         If [prop].GetValue(item) IsNot Nothing Then
            If [prop].GetValue(item).Equals(key) Then
               Return i
            End If
         End If
      Next
      Return -1
   End Function

   ''' <summary>
   ''' Gets the SupportsSortCore flag
   ''' </summary>
   Protected Overloads Overrides ReadOnly Property SupportsSortingCore() As Boolean
      Get
         Return True
      End Get
   End Property

   ''' <summary>
   ''' Gets the IsSortedCore flag
   ''' </summary>
   Protected Overloads Overrides ReadOnly Property IsSortedCore() As Boolean
      Get
         Return mSorted
      End Get
   End Property

   ''' <summary>
   ''' Gets the sort direction
   ''' </summary>
   Protected Overloads Overrides ReadOnly Property SortDirectionCore() As ListSortDirection
      Get
         Return mSortDirection
      End Get
   End Property

   ''' <summary>
   ''' Gets the SortPropertyCore propertydescriptor
   ''' </summary>
   Protected Overloads Overrides ReadOnly Property SortPropertyCore() As PropertyDescriptor
      Get
         Return mSortProperty
      End Get
   End Property

   ''' <summary>
   ''' Removes any applied sort.
   ''' </summary>
   Protected Overloads Overrides Sub RemoveSortCore()
      If mSorted Then
         Clear()
         For Each item As T In mOriginalCollection
            Add(item)
         Next
         mOriginalCollection.Clear()
         mSortProperty = Nothing
         mSortDescriptions = Nothing
         mSorted = False
      End If
   End Sub
#Region "IBindingListView Members"

   ''' <summary>
   ''' Apply a sort
   ''' </summary>
   ''' <param name="sorts"></param>
   Sub ApplySort(ByVal sorts As ListSortDescriptionCollection) Implements System.ComponentModel.IBindingListView.ApplySort
      mSortProperty = Nothing
      mSortDescriptions = sorts
   End Sub

   ''' <summary>
   ''' Get or sets the filter
   ''' </summary>
   ''' <value>Filter string</value>
   Property Filter() As String Implements System.ComponentModel.IBindingListView.Filter
      Get
         Return mFilterString
      End Get
      Set(ByVal value As String)
         mFilterString = value
         mFiltered = True
         UpdateFilter()
      End Set
   End Property

   ''' <summary>
   ''' Remove the filter
   ''' </summary>
   Sub RemoveFilter() Implements System.ComponentModel.IBindingListView.RemoveFilter
      If Not mFiltered Then
         Return
      End If
      mFilterString = Nothing
      mFiltered = False
      mSorted = False
      mSortDescriptions = Nothing
      mSortProperty = Nothing
      Clear()
      For Each item As T In mOriginalCollection
         Add(item)
      Next
      mOriginalCollection.Clear()
   End Sub

   ''' <summary>
   ''' Gets the SortDescriptionCollection
   ''' </summary>
   ReadOnly Property SortDescriptions() As ListSortDescriptionCollection Implements System.ComponentModel.IBindingListView.SortDescriptions
      Get
         Return mSortDescriptions
      End Get
   End Property

   ''' <summary>
   ''' Gets the SupportsAdvancedSorting flag
   ''' </summary>
   ReadOnly Property SupportsAdvancedSorting() As Boolean Implements System.ComponentModel.IBindingListView.SupportsAdvancedSorting
      Get
         Return True
      End Get
   End Property

   ''' <summary>
   ''' Gets the SupportsFiltering flag
   ''' </summary>
   ReadOnly Property SupportsFiltering() As Boolean Implements System.ComponentModel.IBindingListView.SupportsFiltering
      Get
         Return True
      End Get
   End Property
#End Region

   ''' <summary>
   ''' Update the filter
   ''' </summary>
   Protected Overridable Sub UpdateFilter()
      Try
         Dim propNamePos As Integer = 0
         Dim equalPos As Integer = mFilterString.IndexOf("="c)
         Dim notPos As Integer = mFilterString.IndexOf("!"c)
         Dim pos As Integer

         If mOriginalCollection.Count = 0 Then
            mOriginalCollection.AddRange(Me)
         End If

         If equalPos > -1 Then
            pos = equalPos
         Else
            pos = notPos
         End If

         Dim currentCollection As New List(Of T)(Me)
         Dim tempCollection As New List(Of T)
         tempCollection.AddRange(currentCollection)

         While pos > 0
            ' Get property name
            Dim propName As String = mFilterString.Substring(propNamePos, pos - propNamePos).Trim()

            ' Get Filter criteria
            Dim startingTicPos As Integer = pos + 1
            Dim endingTicPos As Integer = mFilterString.IndexOf("'"c, startingTicPos + 1)
            Dim criteria As String = mFilterString.Substring(startingTicPos + 1, endingTicPos - startingTicPos - 1)
            Dim asteriskPos As Integer = criteria.IndexOf("*"c)

            ' Get a property descriptor for the filter property
            Dim propDesc As PropertyDescriptor = TypeDescriptor.GetProperties(GetType(T))(propName)

            For Each item As T In currentCollection
               Dim value As Object = propDesc.GetValue(item)

               Select Case asteriskPos
                  Case Is = 0
                     If criteria.Length > 1 Then
                        If Microsoft.VisualBasic.Right(value.ToString, criteria.Length - 1) <> _
                           Microsoft.VisualBasic.Right(criteria, criteria.Length - 1) Then
                           tempCollection.Remove(item)
                        End If
                     End If
                  Case Is > 0
                     If asteriskPos = criteria.Length - 1 Then
                        If Microsoft.VisualBasic.Left(value.ToString, asteriskPos) <> _
                           Microsoft.VisualBasic.Left(criteria, asteriskPos) Then
                           tempCollection.Remove(item)
                        End If
                     Else
                        If (Microsoft.VisualBasic.Left(value.ToString, asteriskPos) <> _
                            Microsoft.VisualBasic.Left(criteria, asteriskPos)) OrElse _
                           (Microsoft.VisualBasic.Right(value.ToString, criteria.Length - (asteriskPos + 1)) <> _
                            Microsoft.VisualBasic.Right(criteria, criteria.Length - (asteriskPos + 1))) Then
                           tempCollection.Remove(item)
                        End If
                     End If
                  Case Else
                     If notPos > -1 Then
                        If criteria = value.ToString Then tempCollection.Remove(item)
                     Else
                        If criteria <> value.ToString() Then tempCollection.Remove(item)
                     End If
               End Select
            Next

            propNamePos = endingTicPos + 2
            pos = mFilterString.IndexOf("="c, endingTicPos)
         End While

         Clear()

         For Each item As T In tempCollection
            Me.Add(item)
         Next
      Catch ex As Exception
         Throw
      End Try

   End Sub
#Region "IBindingList overrides"

   ''' <summary>
   ''' Gets the AllowNew flag
   ''' </summary>
   Overloads ReadOnly Property AllowNew() As Boolean
      Get
         Return CheckReadOnly()
      End Get
   End Property

   ''' <summary>
   ''' Gets the AllowRemove flag
   ''' </summary>
   Overloads ReadOnly Property AllowRemove() As Boolean
      Get
         Return CheckReadOnly()
      End Get
   End Property

   ''' <summary>
   ''' Gets the CheckReadOnly flag
   ''' </summary>
   Private Function CheckReadOnly() As Boolean
      If mSorted OrElse mFiltered Then
         Return False
      Else
         Return True
      End If
   End Function
#End Region
#Region "IRaiseItemChangedEvents Members"

   ''' <summary>
   ''' Gets the RaisesItemChangeEvents flag
   ''' </summary>
   ReadOnly Property RaisesItemChangedEvents() As Boolean
      Get
         Return True
      End Get
   End Property
#End Region
End Class

